package ga4robot;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Random;
import javax.xml.parsers.ParserConfigurationException;
import org.xml.sax.SAXException;

/**
 * 子代
 * @author 邢兴
 */
public class Generation {
    /**
     * 种群大小
     */
    private int population = 0;
    /**
     * 基因位数
     */
    private int geneCount = 0;
    /**
     * 交叉概率
     */
    private double crossoverPossible = 0;
    /**
     * 变异概率
     */
    private double mutationPossible = 0;
    /**
     * 染色体组
     */
    private ArrayList<Chromosome> chromosomes = null;
    /**
     * 历史最优解
     */
    private Chromosome bestC = null;

    /**
     * 构造函数，加载配置文件，随机初始化基因组
     * @throws java.io.IOException
     */
    public Generation() throws IOException{       
        loadConfig();
        GeneGroup.initGroup(geneCount);        
        init();
    }

    /**
     * 构造函数
     * @param config    基因组输入流
     * @throws java.io.IOException
     * @throws org.xml.sax.SAXException
     * @throws javax.xml.parsers.ParserConfigurationException
     */
    public Generation(InputStream config) throws IOException, SAXException,
            ParserConfigurationException {
        loadConfig();
        GeneGroup.loadGroup(config);
        init();        
    }
    /**
     * 从配置文件中加载数据
     * @throws java.io.IOException
     */
    private void loadConfig() throws IOException{
        population = Config.singlation().getPopulation();
        geneCount = Config.singlation().getGeneCount();
        crossoverPossible = Config.singlation().getCrossoverPossible();
        mutationPossible = Config.singlation().getMutationPossible();
    }
    /**
     * 初始化
     */
    private void init() {
        chromosomes = new ArrayList<Chromosome>(population);
        for(int i = 0; i < population; i++) {
            Chromosome chromosome = new Chromosome(geneCount);
            for(int j = 0; j < geneCount; j ++) {
                chromosome.addGene(j, GeneGroup.getGroup(j).getRandomGene());
            }
            chromosomes.add(chromosome);
        }
    }
    /**
     * 新子代
     */
    public void newGeneration() {
        // 选取进行交叉，越优秀机会越高
        ArrayList<Chromosome> crossPop = selection(crossoverPossible);        
        for(int i = 0; i < crossPop.size(); i ++) {
            if (i == 0) {
                chromosomes.addAll(crossPop.get(0).crossover(crossPop.get(crossPop.size() - 1)));
            } else {
                chromosomes.addAll(crossPop.get(i).crossover(crossPop.get(i - 1)));
            }
        }
        
        // 变异
        Random r = new Random();
        int mutationCount = (int)(chromosomes.size() * mutationPossible);
        for(int i = 0; i < mutationCount; i ++) {
            chromosomes.get(r.nextInt(chromosomes.size())).mutation();
        }
        // 选择
        chromosomes = selection(population);
        Chromosome tmpC = getBest();
        if (bestC == null || bestC.evaluation() < tmpC.evaluation()) {
            bestC = tmpC;
        }
    }
    /**
     *
     * @return 历史最优
     */
    public Chromosome getHistoryBest() {
        return bestC;
    }

    /**
     *
     * @return 当前最优
     */
    public Chromosome getBest() {
        int max = 0;
        for(int i = 1; i < chromosomes.size(); i ++) {
            if (chromosomes.get(max).evaluation() < chromosomes.get(i).evaluation()) {
                max = i;
            }
        }        
        return chromosomes.get(max);
    }

    /**
     * 
     * @return 评估函数合
     */
    private double totalEval() {
        double value = 0;
        for(Chromosome c: chromosomes) {
            value += c.evaluation();
        }
        return value;
    }

    /**
     * 选择
     * @param percent   选择概率
     * @return  染色体组
     */
    private ArrayList<Chromosome> selection(double percent) {
        double total = totalEval();
        double p[] = new double[chromosomes.size()];
        
        for(int i = 0; i < chromosomes.size(); i ++) {
            if (i == 0) {
                p[i] = chromosomes.get(0).evaluation() / total;
            } else {
                p[i] = p[i-1] + chromosomes.get(i).evaluation() / total;
            }
        }
        
        int crossoverCount = (int)(chromosomes.size() * percent);
        Random r = new Random();
        ArrayList<Chromosome> result = new ArrayList<Chromosome>(crossoverCount);
        result.add(getBest());        
        while(result.size() < crossoverCount) {
            double k = r.nextDouble();
            for(int j = chromosomes.size() - 1; j >= 0 ; j --) {
                if (k > p[j]) {
                    result.add(chromosomes.get(j));  
                    break;
                }
            }            
        }
        return result;        
    }
    
    /**
     * 选择
     * @param count 总数
     * @return
     */
    private ArrayList<Chromosome> selection(int count) {
        double total = totalEval();
        double p[] = new double[chromosomes.size()];
        
        for(int i = 0; i < chromosomes.size(); i ++) {
            if (i == 0) {
                p[i] = chromosomes.get(0).evaluation() / total;
            } else {
                p[i] = p[i-1] + chromosomes.get(i).evaluation() / total;
            }
        }
              
        Random r = new Random();        
        ArrayList<Chromosome> result = new ArrayList<Chromosome>(count);
        result.add(getBest());
        while(result.size() < count) {
            double k = r.nextDouble();
            for(int j = chromosomes.size() - 1; j >= 0 ; j --) {
                if (k > p[j]) {
                    result.add(chromosomes.get(j));  
                             
                    break;
                }
            }            
        }
        return result;
    }
}
